import math, time

from smartpray4dpackage.config import SIZE
from smartpray4dpackage.colorhelper import IntToRGB
 

class Renderer:  

    size       = None
    focusZ     = None
    
    initialDir = None

    def __init__(self, size = 100):
    
        self.size       = size
        self.focusZ     = self.size
        self.initialDir = [None]*self.size*self.size*3

        hs = (self.size-1)*0.5  

        #initial spatial direction 
        idx = 0
        for j in xrange(self.size):
            for i in xrange(self.size):
                l = 1/math.sqrt((i - hs)*(i - hs) + (j - hs)*(j - hs) + self.focusZ*self.focusZ);
                self.initialDir[idx] = (i-hs) * l
                idx += 1
                self.initialDir[idx] = (j-hs) * l; 
                idx += 1;
                self.initialDir[idx] = self.focusZ   * l; 
                idx += 1
                
        
    def Render(self, scene, area):
                      
        #clock speed
        t0 = time.clock()
         
        #move light
        scene.light.x = math.cos(scene.time*0.6)*100
        scene.light.y = math.sin(scene.time*1.1)*25+100
        scene.light.z = math.sin(scene.time*0.9)*100-100
        scene.ldir.x  = -scene.light.x
        scene.ldir.y  = -scene.light.y
        scene.ldir.z  = -scene.light.z
        scene.ldir.Normalize()

        #move spheres
        kmax = len(scene.spheres)
        for k in xrange(kmax):
            scene.spheres[k].Update(scene.time)
        #each pixel...
        idx=0 
        for j in xrange(self.size):
            for i in xrange(self.size):
                #offset and direction
                ox = scene.camX;
                oy = scene.camY;
                oz = scene.camZ;
                dx = self.initialDir[idx]
                idx += 1
                dy = self.initialDir[idx]
                idx += 1
                dz = self.initialDir[idx] 
                idx += 1
                pixel = 0
                #4 samples
                for l in xrange(1,5):
                    tmin = 99999
                    hit  = None
                    #walk spheres
                    for k in xrange(kmax):
                        s   = scene.spheres[k]                        
                        dsx = ox - s.x
                        dsy = oy - s.y
                        dsz = oz - s.z
                        B   = dsx * dx + dsy * dy + dsz * dz
                        C   = dsx * dsx + dsy * dsy + dsz * dsz - s.r2
                        D   = B * B - C
                        if D > 0:
                            t = - B - math.sqrt(D)
                            if (t > 0) and (t < tmin):
                                tmin = t
                                hit = s
                    #sphere hit?
                    if hit:
                        ox += dx * tmin
                        oy += dy * tmin
                        oz += dz * tmin
                        nx = ox - hit.x
                        ny = oy - hit.y
                        nz = oz - hit.z
                        n = 1 / math.sqrt(nx*nx + ny*ny + nz*nz)
                        nx *= n
                        ny *= n
                        nz *= n
                        n = -(nx*dx + ny*dy + nz*dz) * 2
                        dx += nx * n
                        dy += ny * n
                        dz += nz * n
                        ln = int((scene.ldir.x * nx + scene.ldir.y * ny + scene.ldir.z * nz) * 256) + 256
                        a = hit.mat.lmap[ln]
                        a >>= scene.refs[l]
                        a &= scene.reff[l]
                        pixel += a
                    else:
                        #background hit?
                        if dy < 0:
                            ln = int((scene.ldir.x * dx + scene.ldir.y * dy + scene.ldir.z * dz) * 256) + 256
                            a  = scene.smap[ln]
                            ln = l - 1
                            a >>= scene.refs[ln]
                            a &= scene.reff[ln]
                            pixel += a
                            break
                        #checkerboard hit!
                        else:
                            tmin = (scene.floorY-oy)/dy
                            ox += dx * tmin
                            oy += dy * tmin
                            oz += dz * tmin
                            dy = -dy
                            ln = dy * 256 + (int(((int(ox+oz)>>7)+(int(ox-oz)>>7))&1)<<9) + 256
                            a = scene.fcol[int(ln)]
                            a >>= scene.refs[l]
                            a &= scene.reff[l]
                            pixel += a
                
                #update userarea bitmap
                area.bmp[i,j] = IntToRGB(pixel)
                area.Redraw(True)
                    
        #update userarea stats
        ms = (time.clock()-t0)*1000
        area.txt = " "+str(1000/ms)[0:5]+" FPS, "+str(ms)[0:7]+ " MSPF"
        area.Redraw(True)
        
        #increment scene time
        scene.time += 0.1